@@ -3,6 +3,7 @@ import os
 import warnings
 from argparse import ArgumentParser
 from itertools import product
+from time import sleep
 from typing import List, Tuple, Optional, Union, Dict
 
 import gym
@@ -12,18 +13,19 @@ from gym.spaces import Discrete, Box
 from gym.wrappers import TimeLimit
 from numpy.random import RandomState
 
-from rlai.actions import Action, DiscretizedAction
+from rlai.actions import Action, DiscretizedAction, ContinuousMultiDimensionalAction
 from rlai.agents.mdp import MdpAgent
-from rlai.environments.mdp import MdpEnvironment
+from rlai.environments.mdp import ContinuousMdpEnvironment
 from rlai.meta import rl_text
+from rlai.models.feature_extraction import NonstationaryFeatureScaler, FeatureExtractor
+from rlai.q_S_A.function_approximation.models.feature_extraction import (
+    StateActionInteractionFeatureExtractor,
+    OneHotCategoricalFeatureInteracter
+)
 from rlai.rewards import Reward
 from rlai.states.mdp import MdpState
-from rlai.utils import parse_arguments
-from rlai.value_estimation.function_approximation.models.feature_extraction import (
-    StateActionInteractionFeatureExtractor,
-    OneHotCategoricalFeatureInteracter,
-    NonstationaryFeatureScaler
-)
+from rlai.utils import parse_arguments, ScatterPlot, ScatterPlotPosition
+from rlai.v_S.function_approximation.models.feature_extraction import StateFeatureExtractor
 
 
 @rl_text(chapter='States', page=1)
@@ -60,7 +62,7 @@ class GymState(MdpState):
 
 
 @rl_text(chapter='Environments', page=1)
-class Gym(MdpEnvironment):
+class Gym(ContinuousMdpEnvironment):
     """
     Generalized Gym environment. Any OpenAI Gym environment can be executed by supplying the appropriate identifier.
     """
@@ -112,6 +114,18 @@ class Gym(MdpEnvironment):
             help='Pass this flag to force the writing of videos into the video directory. Will overwrite/delete content in the directory.'
         )
 
+        parser.add_argument(
+            '--steps-per-second',
+            type=int,
+            help='Number of steps per second when displaying videos.'
+        )
+
+        parser.add_argument(
+            '--plot-environment',
+            action='store_true',
+            help='Pass this flag to plot environment values (e.g., state).'
+        )
+
         return parser
 
     @classmethod
@@ -119,7 +133,7 @@ class Gym(MdpEnvironment):
             cls,
             args: List[str],
             random_state: RandomState
-    ) -> Tuple[MdpEnvironment, List[str]]:
+    ) -> Tuple[ContinuousMdpEnvironment, List[str]]:
         """
         Initialize an environment from arguments.
 
@@ -130,7 +144,7 @@ class Gym(MdpEnvironment):
 
         parsed_args, unparsed_args = parse_arguments(cls, args)
 
-        gym_env = Gym(
+        gym_env = cls(
             random_state=random_state,
             **vars(parsed_args)
         )
@@ -157,22 +171,40 @@ class Gym(MdpEnvironment):
         # map discretized actions back to continuous space
         if isinstance(a, DiscretizedAction):
             gym_action = a.continuous_value
+        # use continuous action values directly (need to be arrays)
+        elif isinstance(a, ContinuousMultiDimensionalAction):
+            gym_action = a.value
+        # use discretized action indices
         else:
             gym_action = a.i
 
         observation, reward, done, _ = self.gym_native.step(action=gym_action)
 
-        if self.render_current_episode:
+        # override reward per environment
+        if self.gym_id == 'LunarLanderContinuous-v2':
+            reward = 2.0 - np.abs(observation[0:5]).sum()
+
+        # call render if rendering manually
+        if self.check_render_current_episode(True):
             self.gym_native.render()
 
-        next_state = GymState(
+        if self.check_render_current_episode(None):
+
+            # sleep if we're restricting steps per second
+            if self.steps_per_second is not None:
+                sleep(1.0 / self.steps_per_second)
+
+            if self.plot_environment:
+                self.state_reward_scatter_plot.update(np.append(observation, reward))
+
+        self.state = GymState(
             environment=self,
             observation=observation,
             terminal=done,
             agent=agent
         )
 
-        return next_state, Reward(i=None, r=reward)
+        return self.state, Reward(i=None, r=reward)
 
     def reset_for_new_run(
             self,
@@ -187,13 +219,13 @@ class Gym(MdpEnvironment):
 
         super().reset_for_new_run(agent)
 
-        # subtract 1 from number of resets in order to render the first episode
-        if self.display_only_rendering:
-            self.render_current_episode = (self.num_resets - 1) % self.render_every_nth_episode == 0
+        if self.plot_environment:
+            self.state_reward_scatter_plot.reset_y_range()
 
         observation = self.gym_native.reset()
 
-        if self.render_current_episode:
+        # call render if rendering manually
+        if self.check_render_current_episode(True):
             self.gym_native.render()
 
         self.state = GymState(
@@ -205,6 +237,30 @@ class Gym(MdpEnvironment):
 
         return self.state
 
+    def check_render_current_episode(
+            self,
+            render_manually: Optional[bool]
+    ) -> bool:
+        """
+        Check whether the current episode is to be rendered.
+
+        :param render_manually: Wether the rendering will be done manually with calls to the render function or automatically
+        as a result of saving videos via the monitor. Pass None to check whether the episode should be rendered,
+        regardless of how the rendering will be done.
+        :return: True if rendered and False otherwise.
+        """
+
+        # subtract 1 from number of resets to render first episode
+        check_result = self.render_every_nth_episode is not None and (self.num_resets - 1) % self.render_every_nth_episode == 0
+
+        if render_manually is not None:
+            if render_manually:
+                check_result = check_result and self.video_directory is None
+            else:
+                check_result = check_result and self.video_directory is not None
+
+        return check_result
+
     def close(
             self
     ):
@@ -214,6 +270,9 @@ class Gym(MdpEnvironment):
 
         self.gym_native.close()
 
+        if self.state_reward_scatter_plot is not None:
+            self.state_reward_scatter_plot.close()
+
     def init_gym_native(
             self
     ) -> Union[EnvSpec, TimeLimit]:
@@ -230,7 +289,7 @@ class Gym(MdpEnvironment):
         # the native gym object uses the max value, so set it to something crazy huge if we're not given a T.
         gym_native._max_episode_steps = 999999999999 if self.T is None else self.T
 
-        # save videos via wrapper
+        # save videos via wrapper if we have a video directory
         if self.render_every_nth_episode is not None and self.video_directory is not None:
             try:
                 gym_native = gym.wrappers.Monitor(
@@ -249,6 +308,74 @@ class Gym(MdpEnvironment):
 
         return gym_native
 
+    def get_state_space_dimensionality(
+            self
+    ) -> int:
+        """
+        Get the dimensionality of the state space.
+
+        :return: Number of dimensions.
+        """
+
+        return self.gym_native.observation_space.shape[0]
+
+    def get_state_dimension_names(
+            self
+    ) -> List[str]:
+        """
+        Get names of state dimensions.
+
+        :return: List of names.
+        """
+
+        if self.gym_id == 'LunarLanderContinuous-v2':
+            names = [
+                'posX',
+                'posY',
+                'velX',
+                'velY',
+                'ang',
+                'angV',
+                'leg1Con',
+                'leg2Con'
+            ]
+        else:  # pragma no cover
+            warnings.warn(f'The state dimension names for {self.gym_id} are unknown. Defaulting to numbers.')
+            names = [str(x) for x in range(0, self.get_state_space_dimensionality())]
+
+        return names
+
+    def get_action_space_dimensionality(
+            self
+    ) -> int:
+        """
+        Get the dimensionality of the action space.
+
+        :return: Number of dimensions.
+        """
+
+        return self.gym_native.action_space.shape[0]
+
+    def get_action_dimension_names(
+            self
+    ) -> List[str]:
+        """
+        Get action names.
+
+        :return: List of names.
+        """
+
+        if self.gym_id == 'LunarLanderContinuous-v2':
+            names = [
+                'main',
+                'side'
+            ]
+        else:  # pragma no cover
+            warnings.warn(f'The action names for {self.gym_id} are unknown. Defaulting to numbers.')
+            names = [str(x) for x in range(0, self.get_action_space_dimensionality())]
+
+        return names
+
     def __init__(
             self,
             random_state: RandomState,
@@ -257,7 +384,9 @@ class Gym(MdpEnvironment):
             continuous_action_discretization_resolution: Optional[float] = None,
             render_every_nth_episode: Optional[int] = None,
             video_directory: Optional[str] = None,
-            force: bool = False
+            force: bool = False,
+            steps_per_second: Optional[int] = None,
+            plot_environment: bool = False
     ):
         """
         Initialize the environment.
@@ -272,6 +401,8 @@ class Gym(MdpEnvironment):
         :param video_directory: Directory in which to store rendered videos.
         :param force: Whether or not to force the writing of videos into the video directory. This will overwrite/delete
         content in the directory.
+        :param steps_per_second: Number of steps per second when displaying videos.
+        :param plot_environment: Whether or not to plot the environment.
         """
 
         super().__init__(
@@ -282,29 +413,28 @@ class Gym(MdpEnvironment):
 
         self.gym_id = gym_id
         self.continuous_action_discretization_resolution = continuous_action_discretization_resolution
-
-        # set up rendering...either display-only or saved videos via wrapper
         self.render_every_nth_episode = render_every_nth_episode
-        self.display_only_rendering = False
-        self.render_current_episode = False
+        if self.render_every_nth_episode is not None and self.render_every_nth_episode <= 0:
+            raise ValueError('render_every_nth_episode must be > 0 if provided.')
+
         self.video_directory = video_directory
-        if self.render_every_nth_episode is not None:
-
-            if self.render_every_nth_episode <= 0:
-                raise ValueError('render_every_nth_episode must be > 0 if provided.')
-
-            # display-only if we don't have a video directory
-            if self.video_directory is None:
-                self.display_only_rendering = True
-                self.render_current_episode = True
-
         self.force = force
+        self.steps_per_second = steps_per_second
+        self.plot_environment = plot_environment
+        self.state_reward_scatter_plot = None
+        if self.plot_environment:
+            self.state_reward_scatter_plot = ScatterPlot(
+                f'{self.gym_id}:  State and Reward',
+                self.get_state_dimension_names() + ['reward'],
+                None
+            )
+
         self.gym_native = self.init_gym_native()
 
         if self.continuous_action_discretization_resolution is not None and not isinstance(self.gym_native.action_space, Box):
             raise ValueError('Continuous-action discretization is only valid for Box action-space environments.')
 
-        # action space is already discrete. initialize n actions from it.
+        # action space is already discrete:  initialize n actions from it.
         if isinstance(self.gym_native.action_space, Discrete):
             self.actions = [
                 Action(
@@ -313,8 +443,20 @@ class Gym(MdpEnvironment):
                 for i in range(self.gym_native.action_space.n)
             ]
 
-        # action space is continuous. discretize it.
-        elif isinstance(self.gym_native.action_space, Box):
+        # action space is continuous and we lack a discretization resolution:  initializee a single, multi-dimensional
+        # action including the min and max values of the dimensions.
+        elif isinstance(self.gym_native.action_space, Box) and self.continuous_action_discretization_resolution is None:
+            self.actions = [
+                ContinuousMultiDimensionalAction(
+                    value=None,
+                    min_values=self.gym_native.action_space.low,
+                    max_values=self.gym_native.action_space.high
+                )
+            ]
+
+        # action space is continuous and we have a discretization resolution:  discretize it. this is generally not a
+        # great approach, as it results in high-dimensional action spaces.
+        elif isinstance(self.gym_native.action_space, Box) and self.continuous_action_discretization_resolution is not None:
 
             box = self.gym_native.action_space
 
@@ -347,16 +489,16 @@ class Gym(MdpEnvironment):
         :return: State dictionary.
         """
 
-        state_dict = dict(self.__dict__)
+        state = dict(self.__dict__)
 
         # the native gym environment cannot be pickled. blank it out.
-        state_dict['gym_native'] = None
+        state['gym_native'] = None
 
-        return state_dict
+        return state
 
     def __setstate__(
             self,
-            state
+            state: Dict
     ):
         """
         Set the state dictionary.
@@ -412,7 +554,11 @@ class CartpoleFeatureExtractor(StateActionInteractionFeatureExtractor):
 
         parsed_args, unparsed_args = parse_arguments(cls, args)
 
-        fex = CartpoleFeatureExtractor(
+        # there shouldn't be anything left
+        if len(vars(parsed_args)) > 0:  # pragma no cover
+            raise ValueError('Parsed args remain. Need to pass to constructor.')
+
+        fex = cls(
             environment=environment
         )
 
@@ -443,7 +589,7 @@ class CartpoleFeatureExtractor(StateActionInteractionFeatureExtractor):
         X = self.feature_scaler.scale_features(X, for_fitting)
 
         contexts = [
-            FeatureContext(
+            CartpoleFeatureContext(
                 cart_left_of_center=state.observation[0] < 0,
                 cart_moving_left=state.observation[1] < 0,
                 pole_left_of_vertical=state.observation[2] < 0,
@@ -492,7 +638,7 @@ class CartpoleFeatureExtractor(StateActionInteractionFeatureExtractor):
         )
 
         self.contexts = [
-            FeatureContext(*context_bools)
+            CartpoleFeatureContext(*context_bools)
             for context_bools in product([True, False], [True, False], [True, False], [True, False])
         ]
 
@@ -505,7 +651,7 @@ class CartpoleFeatureExtractor(StateActionInteractionFeatureExtractor):
         )
 
 
-class FeatureContext:
+class CartpoleFeatureContext:
     """
     Categorical context within with a feature explains an outcome independently of its explanation in another context.
     This works quite similarly to traditional categorical interactions, except that they are specified programmatically
@@ -527,6 +673,7 @@ class FeatureContext:
         :param pole_left_of_vertical: Left of vertical.
         :param pole_rotating_left: Rotating left.
         """
+
         self.cart_left_of_center = cart_left_of_center
         self.cart_moving_left = cart_moving_left
         self.pole_left_of_vertical = pole_left_of_vertical
@@ -544,7 +691,7 @@ class FeatureContext:
         :return: True if equal and False otherwise.
         """
 
-        other: FeatureContext
+        other: CartpoleFeatureContext
 
         return self.id == other.id
 
@@ -582,3 +729,75 @@ class FeatureContext:
         """
 
         return f'{self.cart_left_of_center}_{self.cart_moving_left}_{self.pole_left_of_vertical}_{self.pole_rotating_left}'
+
+
+@rl_text(chapter='Feature Extractors', page=1)
+class ContinuousFeatureExtractor(StateFeatureExtractor):
+    """
+    A feature extractor for continuous OpenAI environments.
+    """
+
+    @classmethod
+    def get_argument_parser(
+            cls
+    ) -> ArgumentParser:
+        """
+        Get argument parser.
+
+        :return: Argument parser.
+        """
+
+        parser = ArgumentParser(
+            prog=f'{cls.__module__}.{cls.__name__}',
+            parents=[super().get_argument_parser()],
+            allow_abbrev=False,
+            add_help=False
+        )
+
+        return parser
+
+    @classmethod
+    def init_from_arguments(
+            cls,
+            args: List[str],
+            environment: Gym
+    ) -> Tuple[FeatureExtractor, List[str]]:
+        """
+        Initialize a feature extractor from arguments.
+
+        :param args: Arguments.
+        :param environment: Environment.
+        :return: 2-tuple of a feature extractor and a list of unparsed arguments.
+        """
+
+        parsed_args, unparsed_args = parse_arguments(cls, args)
+
+        # there shouldn't be anything left
+        if len(vars(parsed_args)) > 0:  # pragma no cover
+            raise ValueError('Parsed args remain. Need to pass to constructor.')
+
+        fex = cls()
+
+        return fex, unparsed_args
+
+    def extract(
+            self,
+            state: GymState,
+    ) -> np.ndarray:
+        """
+        Extract state features.
+
+        :param state: State.
+        :return: State-feature vector.
+        """
+
+        return state.observation
+
+    def __init__(
+            self
+    ):
+        """
+        Initialize the feature extractor.
+        """
+
+        super().__init__()
